
#define ROLL       0
#define PITCH      1
#define YAW        2

#define I2C_SPEED 100000L     //100kHz normal mode, this value must be used for a genuine WMP
//#define I2C_SPEED 400000L   //400kHz fast mode, it works only with some WMP clones

//enable internal I2C pull ups
//#define INTERNAL_I2C_PULLUPS

// for Arduino MEGA
#define I2C_PULLUPS_ENABLE         PORTD |= 1<<0; PORTD |= 1<<1;       // PIN 20&21 (SDA&SCL)
#define I2C_PULLUPS_DISABLE        PORTD &= ~(1<<0); PORTD &= ~(1<<1);


static uint8_t rawADC_NUN[6];
static int16_t accADC[3];


void setup ()
{
  Serial.begin (115200);
  delay(100);

  i2c_init();
  delay(100);

  i2c_ACC_init();

  Serial.println("\nFinished setup");
}


void loop ()
{
  i2c_ACC_getADC();

  Serial.print("\nX: "); 
  Serial.print(rawADC_NUN[2], DEC);
  Serial.print("\tY: "); 
  Serial.print(rawADC_NUN[3], DEC);
  Serial.print("\tZ: "); 
  Serial.print(rawADC_NUN[4], DEC);

  Serial.print("\nROLL: "); 
  Serial.print(accADC[ROLL], DEC);
  Serial.print("\tPITCH: "); 
  Serial.print(accADC[PITCH], DEC);
  Serial.print("\tYAW: "); 
  Serial.print(accADC[YAW], DEC);

  delay (500);
}


void i2c_ACC_init() {
  TWBR = ((16000000L / I2C_SPEED) - 16) / 2;

  // disable encryption
  // look at <http://wiibrew.org/wiki/Wiimote#The_New_Way> at "The New Way"
  i2c_rep_start(0xA4 + 0);//I2C write direction => 0
  i2c_write(0xF0); 
  i2c_write(0x55);
  i2c_rep_start(0xA4 + 0);//I2C write direction => 0
  i2c_write(0xFB); 
  i2c_write(0x00);
  
//  look at <http://wiibrew.org/wiki/Wiimote#The_Old_Way> at "The Old Way"
//  i2c_rep_start(0xA4 + 0);//I2C write direction => 0
//  i2c_write(0x40); 
//  i2c_write(0x00);

  //delay(250);

  // display the identification bytes, must be "00 00 A4 20 00 00" for the Nunchuck
  i2c_rep_start(0xA4 + 0);//I2C write direction => 0
  i2c_write(0xFA);
  i2c_rep_start(0xA4 + 1);//I2C read direction => 1
  for(uint8_t i = 0; i < 5; i++) {
    rawADC_NUN[i]=i2c_readAck();
  }  
  rawADC_NUN[5]= i2c_readNak();

  Serial.print("Ident=");
  for (byte i = 0; i < 6; i++)
  {
    Serial.print(rawADC_NUN[i], HEX);
    Serial.print(' ');
  }
  Serial.println();

}

void i2c_ACC_getADC() {
  //TWBR = ((16000000L / I2C_SPEED) - 16) / 2; // change the I2C clock rate. !! you must check if the nunchuk is ok with this freq

  //for(byte i = 0; i < 3; i++)  {
  i2c_rep_start(0xA4 + 0);//I2C write direction => 0
  i2c_write(0x00);
  //}
  i2c_rep_start(0xA4 + 1);//I2C read direction => 1
  for(uint8_t i = 0; i < 5; i++)
    rawADC_NUN[i]=i2c_readAck();
  rawADC_NUN[5]= i2c_readNak();
  
  accADC[ROLL]  =   ( (rawADC_NUN[3]<<2)        + ((rawADC_NUN[5]>>4)&0x2) );
  accADC[PITCH] = - ( (rawADC_NUN[2]<<2)        + ((rawADC_NUN[5]>>3)&0x2) );
  accADC[YAW]   = - ( ((rawADC_NUN[4]&0xFE)<<2) + ((rawADC_NUN[5]>>5)&0x6) );
}


// *********************
// I2C general functions
// *********************

// Mask prescaler bits : only 5 bits of TWSR defines the status of each I2C request
#define TW_STATUS_MASK	(1<<TWS7) | (1<<TWS6) | (1<<TWS5) | (1<<TWS4) | (1<<TWS3)
#define TW_STATUS       (TWSR & TW_STATUS_MASK)

static uint32_t neutralizeTime;

void i2c_init(void) {
#if defined(INTERNAL_I2C_PULLUPS)
  I2C_PULLUPS_ENABLE
#else
  I2C_PULLUPS_DISABLE
#endif
  TWSR = 0;        // no prescaler => prescaler = 1
  TWBR = ((16000000L / I2C_SPEED) - 16) / 2; // change the I2C clock rate
  TWCR = 1<<TWEN;  // enable twi module, no interrupt
}

void i2c_rep_start(uint8_t address) {
  TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN) | (1<<TWSTO); // send REAPEAT START condition
  waitTransmissionI2C(); // wait until transmission completed
  checkStatusI2C(); // check value of TWI Status Register
  TWDR = address; // send device address
  TWCR = (1<<TWINT) | (1<<TWEN);
  waitTransmissionI2C(); // wail until transmission completed
  checkStatusI2C(); // check value of TWI Status Register
}

void i2c_write(uint8_t data ) {	
  TWDR = data; // send data to the previously addressed device
  TWCR = (1<<TWINT) | (1<<TWEN);
  waitTransmissionI2C(); // wait until transmission completed
  checkStatusI2C(); // check value of TWI Status Register
}

uint8_t i2c_readAck() {
  TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);
  waitTransmissionI2C();
  return TWDR;
}

uint8_t i2c_readNak(void) {
  TWCR = (1<<TWINT) | (1<<TWEN);
  waitTransmissionI2C();
  return TWDR;
}

void waitTransmissionI2C() {
  uint8_t count = 255;
  while (count-->0 && !(TWCR & (1<<TWINT)) );
  if (count<2) { //we are in a blocking state => we don't insist
    TWCR = 0;  //and we force a reset on TWINT register
    neutralizeTime = micros(); //we take a timestamp here to neutralize the value during a short delay after the hard reset
    Serial.println("we are in a blocking state => we don't insist");
  } 
}

void checkStatusI2C() {
  if ( TW_STATUS  == 0xF8) { //TW_NO_INFO : this I2C error status indicates a wrong I2C communication.
    // WMP does not respond anymore => we do a hard reset. I did not find another way to solve it. 
    // It takes only 13ms to reset and init to WMP or WMP+NK
    TWCR = 0;
    //POWERPIN_OFF //switch OFF WMP
    //delay(1);  
    //POWERPIN_ON  //switch ON WMP
    delay(10);
    //#if defined(ITG3200) || defined(L3G4200D)
    //#else
    //i2c_WMP_init(0);
    //#endif
    neutralizeTime = micros(); //we take a timestamp here to neutralize the WMP or WMP+NK values during a short delay after the hard reset

    //Serial.println("WMP does not respond anymore => we do a hard reset.");
  }
}




